require 'json'
require 'rest-client'
require 'postcodeinfo/exceptions'

module PostcodeInfo
  class Client

    attr_accessor :auth_token, :api_uri

    # params:
    #   auth_token: 'your auth token' (optional)
    #   api_url:    'url of a postcodeinfo server' (optional)
    #               If this is not supplied, it will be inferred from the params below
    #               If no value can be inferred, it will default to http://localhost:8000/
    #   env:        (development|test|staging|production) (optional - default is development)
    #               If provided, the API url will be looked up in .api_urls by
    #               this value.
    #               If not provided, it will look for a RAILS_ENV environment variable
    #               and use that.
    #               If it still can't work out the environment, it will default to development
    def initialize(opts={})
      @auth_token = opts[:auth_token]
      @api_uri = URI.parse(infer_api_url(opts))
    end

    def api_url
      @api_uri.to_s
    end

    def api_url=(url)
      @api_uri = URI.parse(url)
    end

    def lookup_postcode(postcode)
      begin
        pc = Postcode.new(postcode, self)
        pc.lookup_info!
        pc.lookup_addresses!
        pc
      rescue RestClient::ResourceNotFound => e
        raise PostcodeInfo::UnrecognisedPostcode.new(e, nil)
      end
    end

    def valid?(postcode)
      begin
        response = make_request('/postcodes/' + postcode)
        case response.code
          when 200 then true
          when /^5[0-9]{2}/ then nil
          when 404 then false
        end

      rescue RestClient::ResourceNotFound => e
        false
      end
    end

    def addresses(postcode)
      response = make_request('/addresses/?postcode=' + postcode)
      handle_response(response)
    end

    def info(postcode)
      response = make_request('/postcodes/' + postcode)
      handle_response(response)
    end

    def make_request(endpoint)
      headers = {
        'Authorization' => 'Token ' + @auth_token
      }
      begin
        response = RestClient::Request.execute( method: :get, 
            url: URI.join(@api_uri, endpoint).to_s, 
            headers: headers, 
            timeout: 5 
          )

      rescue Exception => ex
        raise wrapped_exception(ex, response)
      end
    end

    protected

      def wrapped_exception(exception, response)
        case exception
        when RestClient::Unauthorized
          PostcodeInfo::InvalidAuthToken.new(exception, response)
        when RestClient::InternalServerError
          PostcodeInfo::ServerError.new(exception, response)
        when SocketError
          PostcodeInfo::ServiceUnavailable.new(exception, response)
        else
          exception
        end
      end

      def handle_response(response)
        JSON.parse(response.body, symbolize_names: true)
      end

      def infer_api_url(opts={})
        opts[:api_url] ||
          from_env( effective_env(opts) )
      end

      def effective_env(opts)
        opts[:env] || ENV['RAILS_ENV'] || :development
      end

      def from_env(env)
        self.class.api_urls[env.to_sym] || self.class.api_urls[:development]
      end

      def self.api_urls
        {
          development: 'http://localhost:8000/',
          test: 'http://localhost:8000/',
          staging: 'https://postcodeinfo-staging.dsd.io/',
          production: 'https://postcodeinfo.service.justice.gov.uk/'
        }
      end
  end
end